from collections import OrderedDict
import json
from pathlib import Path
import shutil

import pandas as pd

from .py3 import HttpClient, NonSuccessResponseError


class FileBatchRunnerClient(HttpClient):
    def __init__(self, conf_json: str, debug=False, max_count=None, **kwargs):
        """
        Iterate a sequence of requests on files of multiple directories
        :param conf_json: path to configuration JSON file
        :param debug: interrupt execution if a non-200 request is received, otherwise continue
        :param max_count: if specified, only process the specified number of input files
        """
        self.debug = debug
        self.conf_root = Path(conf_json).parent
        with open(conf_json, 'r') as fh:
            self.conf = json.load(fh)

        self.paths = {
            'conf': Path(self.conf['paths']['conf']),
            'input': Path(self.conf['paths']['input']),
            'output': Path(self.conf['paths']['output']),
        }
        for pa in self.paths.values():
            pa.mkdir(parents=True, exist_ok=True)

        shutil.copy(conf_json, self.paths['output'] / conf_json.name)

        self.stacks = self.get_stacks(max_count=max_count)

        self.tasks = {}

        self.write_df()

        if not self.stacks['exists'].all():
            raise FileNotFoundError(f'Found non-existent files, described in {self.pa_csv}')
        return super().__init__(**kwargs)

    def message(self, message):
        print(message)

    def write_df(self):
        for rt in [self.conf_root, self.paths['output']]:
            pa = rt / 'filelist.csv'
            self.stacks.to_csv(pa)
            self.message(f'Wrote stacks table to {pa}.')

    def hit(self, method, endpoint, params=None, body=None, catch=True, **kwargs):
        resp = super(FileBatchRunnerClient, self).hit(method, endpoint, params=params, body=body)
        if resp.status_code != 200:
            self.message(f'Non-200 response from {endpoint}:\n{resp.text}')
            if not catch:
                raise NonSuccessResponseError(resp.text)
        return resp.json()

    def restart_session(self):
        self.hit('get', 'session/restart')

    def verify_server(self):
        try:
            # test server communication
            self.get('')
        except Exception as e:
            self.message('Could not find server at: ' + self.uri)
            raise(e)
        self.message('Verified server is online at: ' + self.uri)

    def watch_path(self, key, path, make=True, verify=False):
        if make:
            path.mkdir(parents=True, exist_ok=True)

        touch_uuid = self.hit(
            'put',
            f'/paths/watch_{key}',
            params={'path': path.__str__(), 'touch': verify}
        )
        if verify:
            pa_touch = path / 'svlt.touch'
            try:
                with open(pa_touch, 'r') as fh:
                    cont = fh.read()
                    assert touch_uuid == cont
                pa_touch.unlink()
            except Exception as e:
                raise WatchPathVerificationError(e)
        self.message(f'Watching {path} for {key} data')

    def setup(self):
        self.watch_path('conf', self.paths['conf'], verify=True, make=False)
        self.watch_path('output', self.paths['output'], verify=True, make=True)
        self.watch_path('input', self.paths['input'], verify=False, make=False)

        for v in self.conf['setup']:
            resp = self.hit(**v, catch=False)
            self.message(f'Completed setup call to ' + v['endpoint'])

    def get_stacks(self, max_count=None):
        paths = []
        for inp in self.conf['inputs']:
            loc = Path(self.paths['input']) / inp['directory']

            # get explicit filenames
            files = inp.get('files', [])

            # get files by pattern
            if pattern := inp.get('pattern'):
                if pattern == '':
                    break
                for f in list(loc.iterdir()):
                    if pattern.upper() in f.name.upper() and f.name not in files:
                        files.append(f.name)
            is_multiposition = inp.get('multiposition', False)
            paths = paths + [{'path': loc / f, 'is_multiposition': is_multiposition} for f in files]
        if max_count is not None:
            df = pd.DataFrame(paths).head(min(max_count, len(paths)))
        else:
            df = pd.DataFrame(paths)
        if len(df) == 0:
            raise EmptyFileListError('No files were found')
        df['exists'] = df['path'].apply(lambda x: x.exists())
        df['parent'] = df['path'].apply(lambda x: x.parent)
        df['filename'] = df['path'].apply(lambda x: x.name)
        df['accessor_id'] = None
        self.message(f'Found {len(df)} input files.')
        return df

    def read_accs(self):
        def _read(row):
            try:
                if row.is_multiposition:
                    return self.put(f'accessors/read_multiposition_file/{row.filename}', query={'lazy': True}).json()
                else:
                    return [self.put(f'accessors/read_from_file/{row.filename}', query={'lazy': True}).json()]
            except NonSuccessResponseError:
                return None

        accessor_ids = []
        for pa_dir, df_gb in self.stacks.groupby('parent'):
            self.watch_path('input', self.paths['input'] / pa_dir, verify=False, make=False)
            df_gb['accessor_id'] = df_gb.apply(_read, axis=1)
            df_gb['position'] = df_gb['accessor_id'].apply(lambda x: range(0, len(x)))
            df_gb = df_gb.explode(['accessor_id', 'position'])
            accessor_ids.append(df_gb)
        self.stacks = pd.concat(accessor_ids).reset_index()
        self.write_df()

    def queue_tasks(self):
        def _task(input_acc_id):
            tasks = self.conf['analyze']
            output_acc_ids = OrderedDict({'input': input_acc_id})
            for v in tasks:
                v['body']['schedule'] = True
                if aim := v.get('accessor_id_mapping'):  # explicitly map inputs to review tasks' outputs
                    for pki, tki in aim.items():
                        v['body'][pki] = output_acc_ids[tki]
                else:  # input is the output of the last task
                    v['body']['accessor_id'] = list(output_acc_ids.values())[-1]
                task_id = self.hit(**v, catch=not self.debug)['task_id']
                v['task_id'] = task_id

                # relate each task ID to its starting input accessor ID and task
                self.tasks[v['task_id']] = {
                    'stack_index': pd.Index(self.stacks['accessor_id']).get_loc(input_acc_id),
                    'task_key': v['key'],
                }
                output_acc_ids[v['key']] = self.hit('get', f'tasks/get_output_accessor/{task_id}')
        self.stacks['accessor_id'].apply(_task)

    def run_tasks(self):
        while task_id := self.hit('get', 'tasks/next'):
            task_info = self.tasks[task_id]
            task_key = task_info['task_key']
            stack_index = task_info['stack_index']
            p = {
                'write_all': True,
                'prefix': f'stack-{stack_index:04d}-task-{task_key}',
                'output_subdirectory': task_key,
            }
            resp = self.hit('put', f'tasks/run/{task_id}', params=p, catch=not self.debug)
            dt = sum(resp['timer'].values())
            self.message(f'Ran task {task_id} in {dt:.3f} sec')

            self.hit('put', f'tasks/delete_accessors/{task_id}')

    def cleanup(self):
        waiting_tasks = self.hit('get', 'tasks/waiting')

        if len(waiting_tasks) > 0:
            raise TasksRemainingError(f'There are tasks that did not run:\n' + str(waiting_tasks.keys()))

    def run(self):
        self.restart_session()
        self.verify_server()
        self.read_accs()
        self.setup()
        self.queue_tasks()
        self.run_tasks()
        self.cleanup()
        self.message('Finished')

class Error(Exception):
    pass

class FileNotFoundError(Error):
    pass

class EmptyFileListError(Error):
    pass

class WatchPathVerificationError(Error):
    pass

class TasksRemainingError(Error):
    pass